// created by Josh Heyde
// last updated 21/03/09

#include "C_FlockingPredator.h"
#include "C_Visualisation.h"

#include <assert.h>

// constructor
C_FlockingPredator::C_FlockingPredator(const int &newGraphicID, const int &newCurrentFrame, const float &newX, const float &newY,
									   const int &newSpeed, const int &newRepulsionRange, const int &newVisionRange, const int &newMaxVelocity)
									   : C_WorldEntity(newGraphicID, newCurrentFrame, newX, newY, newSpeed, newMaxVelocity, newRepulsionRange, newVisionRange, C_WorldEntity::ePredator)
{
}

// destructor
C_FlockingPredator::~C_FlockingPredator()
{
}

// member functions

// the collision result of this entity
void C_FlockingPredator::collisionResult()
{
}

// determines the direction the entity will move in
void C_FlockingPredator::determineDirection(std::vector <C_WorldEntity*> v_entityVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(!v_entityVector.empty());

	// resets the direction
	desiredDirection->x = 0;
	desiredDirection->y = 0;

	// models the cohesion rule
	*desiredDirection += cohesion(v_entityVector);
	// models the separation rule
	*desiredDirection += avoidance(v_entityVector);
	// models the aligment rule
	*desiredDirection += alignment(v_entityVector);

	// only calculates a new direction if the desired direction isn't null (0, 0)
	// or the same as the current direction
	if(desiredDirection->x != 0.0f && desiredDirection->y != 0.0f && *desiredDirection != *direction)
	{
		*direction = *desiredDirection;
	}

	// ensures the entity doesn't move too fast
	controlVelocity();
}

// determines whether the entity passed in is an enemy or not
bool C_FlockingPredator::isEnemyOf(C_WorldEntity *other) const
{
	if(other != this)
	{
		return(other->getType() == C_WorldEntity::ePrey);
	}
	else
	{
		return false;
	}
}

// draws the sprite
void C_FlockingPredator::render(const float &currentTime)
{
	// error handling statements
	// asserting certain statements for this function
	assert(currentTime >= 0);
	
	C_Vector inter(currentPosition->x, currentPosition->y);

	// only interpolates if the next position is different to the current position
	if((currentPosition->x != nextPosition->x) || (currentPosition->y != nextPosition->y))
	{
		inter.x += currentTime * (nextPosition->x - currentPosition->x);
		inter.y += currentTime * (nextPosition->y - currentPosition->y);
	}

	VIS.drawSprite(graphicID, (int) inter.x, (int) inter.y, currentFrame);
}

// updates the entity
void C_FlockingPredator::update()
{
	C_WorldEntity::update();

	// prevents the entity going off screen
	if(nextPosition->x > VIS.getScreenWidth())
	{
		direction->x = (float)-maxVelocity;
	}

	if(nextPosition->x < 0)
	{
		direction->x = (float)maxVelocity;
	}

	if(nextPosition->y > VIS.getScreenHeight())
	{
		direction->y = (float)-maxVelocity;
	}

	if(nextPosition->y < 0)
	{
		direction->y = (float)maxVelocity;
	}

	// moves the entity
	*nextPosition += (*direction * (float)speed);
}


////////////////////////
// behavioural functions
////////////////////////

// seeks to align the direction of this entity with those surrounding it
C_Vector C_FlockingPredator::alignment(std::vector<C_WorldEntity*> v_entityVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(!v_entityVector.empty());

	// creates a temporary vector to work out the direction to separate
	C_Vector temp(0, 0);

	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		// checks for a comparison against itself
		if(v_entityVector[i] != this)
		{
			// checks that the entity to move towards is active
			if(v_entityVector[i]->getActive())
			{
				// checks to see if it is 'friendly'
				if(!isEnemyOf(v_entityVector[i]))
				{
					// checks to see if the target is in range
					if(inRange(v_entityVector[i], visionRange))
					{	
						// adds up all the vectors between this entity and
						// the surrounding entities
						temp += *v_entityVector[i]->getDirection();
					}
				}
			}
		}
	}

	return temp;
}

// seeks to avoid entities that are too close to this entity
C_Vector C_FlockingPredator::avoidance(std::vector<C_WorldEntity*> v_entityVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(!v_entityVector.empty());

	// creates a temporary vector to work out the direction to separate
	C_Vector temp(0, 0);

	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		if(v_entityVector[i] != this)
		{
			// checks that the entity to move towards is active
			if(v_entityVector[i]->getActive())
			{
				if(!isEnemyOf(v_entityVector[i]))
				{
					// adds up all the vectors between this entity and
					// the surrounding entities
					if(inRange(v_entityVector[i], repulsionRange))
					{
						temp += (*currentPosition - *v_entityVector[i]->getCurrentPosition());
					}
				}
			}
		}
	}

	return temp;
}

// seeks to aim towards the average position of the surrounding entities
C_Vector C_FlockingPredator::cohesion(std::vector<C_WorldEntity*> v_entityVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(!v_entityVector.empty());

	// creates a temporary vector to work out the direction
	C_Vector temp(0, 0);
	int count = 0;

	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		// checks for a comparison against itself
		if(v_entityVector[i] != this)
		{
			// checks that the entity to move towards is active
			if(v_entityVector[i]->getActive())
			{
				// checks to see if the entity is 'friendly'
				if(inRange(v_entityVector[i], visionRange))
				{
					// adds up all the current positions of the surrounding entities
					temp += *v_entityVector[i]->getCurrentPosition();
					count++;
				}
			}
		}
	}

	// sets the entity to aim for the average position of the nearby entities
	if(count > 0)
	{
		temp /= ((float) count);
		temp -= *currentPosition;
	}
	else
	{
		temp.x = 0;
		temp.y = 0;
	}

	return temp;
}
