// created by Josh Heyde
// last updated 25/01/09

#include "C_World.h"
#include "C_WorldEntity.h"
#include "C_FlockingEntity.h"
#include "C_FlockingPredator.h"
#include "C_Visualisation.h"
#include "C_UI.h"
#include <HAPI_lib.h>
#include <assert.h>

#include <sstream>
#include <ctime>

using std::string;

// initialisation of the singleton
C_World* C_World::world = NULL;

// creation of a singleton
void C_World::createWorld()
{
	if(!world)
		world = new C_World();
}

// deletion of a singleton
void C_World::deleteWorld()
{
	// error handling statements
	// asserting certain statements for this function
	assert(world != NULL);

	delete world;
}

// constructor
C_World::C_World() : currentTime(0), previousTime(0), DELAY(50)
{
	// setting up the screen size
	 const int screenWidth = 800;
     const int screenHeight = 600;

	// initialises the visualisation singleton
	VIS.createVisual(screenWidth, screenHeight);

	// creates the UI singleton
	INPUT.createInput();

	srand((unsigned int)time(0));
}

// destructor
C_World::~C_World()
{
	// deletes all created world entities
	for (std::vector<C_WorldEntity*>::const_iterator iter = v_entityVector.begin(); iter < v_entityVector.end(); iter++)
		delete *iter;

	// deletes the visualisation singleton
	VIS.deleteVisual();

	// deletes the UI singleton
	INPUT.deleteInput();
}

// other member functions
bool C_World::initialise()
{
	// checking that the program initialises correctly
	if(!VIS.initialise())
	{
		// deletes all created world entities
		for (std::vector<C_WorldEntity*>::const_iterator iter = v_entityVector.begin(); iter < v_entityVector.end(); iter++)
			delete *iter;
		VIS.deleteVisual();
		INPUT.deleteInput();

		return false;
	}

	return true;
}

void C_World::loadData()
{
	// temporary variables for use in construction
	int ID = 0;
	float x = 0;
	float y = 0;
	int width = 16;
	int height = 16;

	const int NUM_OF_PREY = 50;
	const int NUM_OF_PREDATORS = 5;

	if(!VIS.createSprite("playerBullet.png", ID, 0, width, height))
	{
		return;
	}

	// creating the prey
	for(int i = 0; i < NUM_OF_PREY; i++)
	{
		x = (float)(rand() % (VIS.getScreenWidth() - width));
		y = (float)(rand() % (VIS.getScreenHeight() - height));

		v_entityVector.push_back(new C_FlockingEntity(ID, 0, x, y, 2, 30, 100, 5));
	}

	if(!VIS.createSprite("enemyBullet.png", ID, 0, width, height))
	{
		return;
	}

	// creating the predators
	for(int i = 0; i < NUM_OF_PREDATORS; i++)
	{
		x = (float)(rand() % (VIS.getScreenWidth() - width));
		y = (float)(rand() % (VIS.getScreenHeight() - height));

		v_entityVector.push_back(new C_FlockingPredator(ID, 0, x, y, 2, 30, 150, 5));
	}
}

template <class T>
std::string toString(const T & t)
{
    std::ostringstream oss; // create a stream
    oss << t;              // insert value to stream
    return oss.str();      // return as a string
}

// calls the UI component to get input
void C_World::getInput()
{
	// refreshes the input devices
	INPUT.refreshKeyboard();
	INPUT.refreshMouse();

	// closes HAPI if esc is pressed
	if(INPUT.isEscPressed())
	{
		HAPI->Close();
	}
}

// renders the sprites of the world entities
void C_World::draw(const float &currentTick)
{
	// refreshes the screen by clearing to black
	VIS.clearToColour(0, 0, 0);

	// rendering each entity
	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		if(v_entityVector[i]->getActive())
		{
			v_entityVector[i]->render(currentTick);
		}
	}

	// displays useful info on screen
	VIS.displayText(0, 13, "Press Esc to Exit", 255, 255, 255);	
}

void C_World::update()
{
	currentTime = (float) HAPI->GetTime();

	// gets input from the user
	getInput();

	// only updates the world entities after a set period of time
	if(currentTime > (previousTime + DELAY))
	{
		// updates every entity
		for(size_t i = 0; i < v_entityVector.size(); i++)
		{
			// only updates the entity if it is active
			if(v_entityVector[i]->getActive())
			{
				// evaluates the direction of every entity
				v_entityVector[i]->determineDirection(v_entityVector);
				
				v_entityVector[i]->update();
			}
		}

		previousTime = currentTime;
	}

	// loops through every entity and checks for collisions against
	// all other entities (optimisations included within the function)
	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		// only checks for collisions if the entity is active
		if(v_entityVector[i]->getActive())
		{
			for(size_t j = i; j < v_entityVector.size(); j++)
			{
				if(v_entityVector[j]->getActive())
				{
					if(v_entityVector[i]->checkCollision(v_entityVector[j]))
					{
						// calls the collision result function for both entities
						v_entityVector[i]->collisionResult();
						v_entityVector[j]->collisionResult();
					}
				}
			}
		}
	}

	draw((float)(currentTime - previousTime) / DELAY);	
}

