// created by Josh Heyde
// last updated 21/03/09

#include "C_WorldEntity.h"
#include "C_Visualisation.h"
#include "C_Vector.h"
#include "C_Rectangle.h"
#include <assert.h>
#include <cmath>

// defines null rather than including the entire cstdlib header file
#define NULL 0

C_WorldEntity::C_WorldEntity(const int &ID, const int &cF, const float &pX, const float &pY, const int &s, const int &mV,
							 const int &rR, const int &vR, const eType &newType)
: active(true), graphicID(ID), currentFrame(cF), speed(s), maxVelocity(mV), repulsionRange(rR), visionRange(vR), type(newType)
{
	turnSpeed = 15;

	currentPosition = new C_Vector(pX, pY);
	nextPosition = new C_Vector(pX, pY);
	direction = new C_Vector(1, 1);
	desiredDirection = new C_Vector(0, 0);

	boundaries = new C_Rectangle(0, 16, 0, 16);
}

C_WorldEntity::~C_WorldEntity()
{
	delete currentPosition;
	delete nextPosition;
	delete direction;
	delete desiredDirection;
	delete boundaries;
}

// mutator functions
void C_WorldEntity::setActive(const bool &newActive)
{
	active = newActive;
}

void C_WorldEntity::setGraphicID(const int &newGraphicID)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newGraphicID >= 0);

	graphicID = newGraphicID;
}

void C_WorldEntity::setSpeed(const int &newSpeed)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newSpeed >= 0);
	
	speed = newSpeed;
}

void C_WorldEntity::setMaxVelocity(const int &newMaxVelocity)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newMaxVelocity <= 0);

	maxVelocity = newMaxVelocity;
}

void C_WorldEntity::setRepulsionRange(const int &newRepulsionRange)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newRepulsionRange <= 0);

	repulsionRange = newRepulsionRange;
}

void C_WorldEntity::setVisionRange(const int &newVisionRange)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newVisionRange <= 0);

	visionRange = newVisionRange;
}

void C_WorldEntity::setTurnSpeed(const int &newTurnSpeed)
{
	turnSpeed = newTurnSpeed;
}

void C_WorldEntity::setCurrentPosition(const int &newX, const int &newY)
{
	currentPosition->x = (float) newX;
	currentPosition->y = (float) newY;
}

void C_WorldEntity::setCurrentPosition(const C_Vector *newVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newVector != NULL);
	
	if(currentPosition != newVector)
	{
		currentPosition->x = newVector->x;
		currentPosition->y = newVector->y;
	}
}

void C_WorldEntity::setNextPosition(const int &newX, const int &newY)
{
	nextPosition->x = (float) newX;
	nextPosition->y = (float) newY;
}

void C_WorldEntity::setNextPosition(const C_Vector *newVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newVector != NULL);
	
	if(nextPosition != newVector)
	{
		nextPosition->x = newVector->x;
		nextPosition->y = newVector->y;
	}
}

void C_WorldEntity::setDirection(const int &newX, const int &newY)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newX < 0);
	assert(newY < 0);

	direction->x = (float) newX;
	direction->y = (float) newY;
}

void C_WorldEntity::setDirection(const C_Vector* newDirection)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newDirection != NULL);

	*direction = *newDirection;
}

void C_WorldEntity::setDesiredDirection(const int &newX, const int &newY)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newX < 0);
	assert(newY < 0);

	desiredDirection->x = (float) newX;
	desiredDirection->y = (float) newY;
}

void C_WorldEntity::setDesiredDirection(const C_Vector* newDesiredDirection)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newDesiredDirection != NULL);

	*desiredDirection = *newDesiredDirection;
}

void C_WorldEntity::setBoundaries(const int &newLeft, const int &newRight, const int &newTop, const int &newBottom)
{
	boundaries->left = newLeft;
	boundaries->right = newRight;
	boundaries->top = newTop;
	boundaries->bottom = newBottom;
}

void C_WorldEntity::setBoundaries(const C_Rectangle *newRect)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newRect != NULL);
	
	if(!(boundaries == newRect))
	{
		boundaries->left = newRect->left;
		boundaries->right = newRect->right;
		boundaries->top = newRect->top;
		boundaries->bottom = newRect->bottom;
	}
}

void C_WorldEntity::setType(const eType &newType)
{
	type = newType;
}

// other member functio

bool C_WorldEntity::checkCollision(C_WorldEntity *other)
{
	// error handling statements
	// asserting certain statements for this function
	assert(other != NULL);
	
	// ensures that the object being compared to is not this object
	if(other != this)
	{
		// only checks for collisions between enemies
		if(isEnemyOf(other))
		{
			// checks for collisions within the rectangle itself
			return (((nextPosition->x <= other->nextPosition->x + other->boundaries->getWidth()) &&
				(nextPosition->x + boundaries->getWidth() >= other->nextPosition->x)) &&
				((nextPosition->y <= other->nextPosition->y + other->boundaries->getHeight()) &&
				(nextPosition->y + boundaries->getHeight() >= other->nextPosition->y)));
		}
	}
	return false;
}

// controls the velocity of the entity
void C_WorldEntity::controlVelocity()
{
	if((fabs(direction->x) > maxVelocity) || (fabs(direction->y) > maxVelocity))
	{
		direction->Normalise();
		*direction *= (float) maxVelocity;
	}
}

// checks to see if an entity is in range
bool C_WorldEntity::inRange(const C_WorldEntity *other, int range) const
{
	// checks to see if a specified entity is within a specified range
	// of this entity
	return(((*currentPosition - *other->getCurrentPosition()).Length()) < range);
}

void C_WorldEntity::update()
{
	*currentPosition = *nextPosition;
}